# -*- coding: utf-8 -*-
# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-celerytask views/forms/actions/components for web ui"""

import logging
import json

import six
from logilab.mtconverter import xml_escape


from cubicweb import tags
from cubicweb.entities.adapters import ITreeAdapter
from cubicweb.view import EntityView
from cubicweb.predicates import is_instance, match_view, adaptable
from cubicweb.web.views import uicfg, tabs
from cubicweb.web.views.cwsources import LogTable, LogTableLayout
from cubicweb.web.views.json import JsonMixIn

from cubicweb_celerytask import STATES

_ = six.text_type


_pvs = uicfg.primaryview_section
_pvs.tag_attribute(('CeleryTask', 'task_name'), 'hidden')
_pvs.tag_object_of(('CeleryTask', 'parent_task', '*'), 'hidden')


class CeleryTaskITreeAdapter(ITreeAdapter):
    __select__ = ITreeAdapter.__select__ & is_instance('CeleryTask')
    tree_relation = 'parent_task'
    child_role = 'subject'
    parent_role = 'object'


class CeleryTaskTabbedPrimaryView(tabs.TabbedPrimaryView):
    """Tabbed primary view for CeleryTask"""
    __select__ = tabs.TabbedPrimaryView.__select__ & is_instance('CeleryTask')
    tabs = [
        _('celerytask.task_general_information'),
        _('celerytask.task_logs'),
    ]
    default_tab = 'celerytask.task_general_information'


class CeleryTaskPrimaryTab(tabs.PrimaryTab):
    """Main tab for CeleryTask"""
    __regid__ = 'celerytask.task_general_information'
    __select__ = EntityView.__select__ & is_instance('CeleryTask')

    def entity_call(self, entity):
        self._cw.add_js(('cubicweb.htmlhelpers.js', 'cubicweb.ajax.js',
                         'cubes.celerytask.js'))
        self._cw.add_onload('cw.celerytask.autorefreshprimary()')
        if not entity.cw_adapt_to('ICeleryTask').finished:
            entity.view('celerytask.task_progress_bar', w=self.w)
        super(CeleryTaskPrimaryTab, self).entity_call(entity)

        if entity.reverse_parent_task:
            self.wview('treeview', rset=entity.related(
                'parent_task', role='object'),
                       subvid='outofcontext')


class CeleryTaskLogs(tabs.TabsMixin, EntityView):
    """View for CeleryTask logs, only displayed if there are some"""
    __regid__ = 'celerytask.task_logs'
    __select__ = EntityView.__select__ & is_instance('CeleryTask')
    default_level = 'Debug'

    def entity_call(self, entity):
        logs = entity.cw_adapt_to('ICeleryTask').logs
        if logs:
            logs = xml_escape(logs).decode('utf-8', errors='ignore')
            self._cw.view('celerytask.task_logs.table',
                          pyvalue=log_to_table(logs), paginate=True,
                          default_level=self.default_level, w=self.w)


class CeleryTaskLogTableLayout(LogTableLayout):
    __select__ = match_view('celerytask.task_logs.table')


class CeleryTaskLogTable(LogTable):
    __regid__ = 'celerytask.task_logs.table'
    headers = [_('severity'), _('date'), _('time'), _('message')]


SEVERITY_LVL = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARNING': logging.WARNING,
    'ERROR': logging.ERROR,
    'FATAL': logging.FATAL,
    'CRITICAL': logging.CRITICAL,
}


def log_to_table(logs):
    rows = []
    for line in logs.splitlines():
        line = line.strip()
        if not line:
            continue
        try:
            severity, date, time, info = line.split(None, 3)
            severity = SEVERITY_LVL[severity]
            rows.append([severity, date, time, info])
        except (ValueError, KeyError):
            if rows:
                rows[-1][-1] += '\n' + line
            else:
                rows.append([logging.INFO, '', '', line])
    return rows


class CeleryTaskProgressBarView(EntityView):
    __select__ = EntityView.__select__ & adaptable('ICeleryTask')
    __regid__ = 'celerytask.task_progress_bar'

    def entity_call(self, entity, **kwargs):
        adapted = entity.cw_adapt_to('ICeleryTask')
        if not adapted.finished:
            self._cw.add_js(('cubicweb.htmlhelpers.js', 'cubicweb.ajax.js',
                             'cubes.celerytask.js'))
            self._cw.add_onload(
                'cw.celerytask.autorefreshprogress(%s, "%s");'
                % (entity.eid, adapted.state))
        progratio = adapted.progress
        if isinstance(progratio, (list, tuple)):
            step, total = progratio
            progress = u'%.0f %%' % (100. * step / total)
        else:
            step = progratio
            total = 1.0
            progress = u'%.0f %%' % (100. * float(progratio))
        self.w(
            tags.tag('progress')(progress,
                                 min=u'0', max=six.text_type(total),
                                 value=six.text_type(step),
                                 id=u'js-cw-celerytask-%s' % entity.eid))


class CeleryTaskJsonView(JsonMixIn, EntityView):
    __select__ = EntityView.__select__ & adaptable('ICeleryTask')
    __regid__ = 'celerytask.jsonexport'

    def entity_call(self, entity, **kwargs):
        adapted = entity.cw_adapt_to('ICeleryTask')
        state = adapted.state
        if state == STATES.PENDING:
            progress = None
            total = 1
        elif adapted.finished:
            progress = 1
            total = 1
        else:
            progress = adapted.progress
            if isinstance(progress, (list, tuple)):
                progress, total = progress
            else:
                total = 1
        result = {'eid': entity.eid,
                  'task_id': adapted.task_id,
                  'task_name': adapted.task_name,
                  'state': state,
                  'progress': progress,
                  'total': total,
                  }
        self.wdata(result)


class CeleryTaskInContextView(EntityView):
    __regid__ = 'incontext'
    __select__ = EntityView.__select__ & adaptable('ICeleryTask')

    def entity_content(self, entity, state):
        return self._cw._(state)

    def entity_call(self, entity):
        adapted = entity.cw_adapt_to('ICeleryTask')
        state, finished = adapted.state, adapted.finished
        html = [tags.a(self.entity_content(entity, state),
                       klass='celerytask-%s' % state.lower(),
                       href=entity.absolute_url())]
        if not finished:
            self._cw.add_js(('cubicweb.htmlhelpers.js', 'cubicweb.ajax.js',
                             'cubes.celerytask.js'))
            self._cw.add_onload("cw.celerytask.setup(%s, %s, '%s');" % (
                entity.eid, json.dumps(state), self.__regid__))
            html.append(entity.view('celerytask.task_progress_bar', w=None))
        self.w(tags.div('\n'.join(html),
                        id='js-celerytask-%s' % entity.eid))


class CeleryTaskOutOfContextView(CeleryTaskInContextView):
    __regid__ = 'outofcontext'
    __select__ = EntityView.__select__ & adaptable('ICeleryTask')

    def entity_content(self, entity, state):
        return entity.dc_long_title()



"""cubicweb-celerytask application package

Run and monitor celery tasks
"""

from six import text_type
_ = text_type


class _States(object):
    """ Helper class to create customizable state "lists". Each state is
    accessible as an attribute and the list may be customized by patching the
    resulting object.
    """
    def __init__(self, *args):
        super(_States, self).__init__()
        for arg in args:
            setattr(self, arg, arg)

    def __contains__(self, item):
        return hasattr(self, item)


STATES = _States(_('PENDING'), _('STARTED'), _('SUCCESS'),
                 _('FAILURE'), _('REVOKED'))


FINAL_STATES = _States(_('SUCCESS'), _('FAILURE'), _('REVOKED'))



# -*- coding: utf-8 -*-
# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import datetime
import json
import logging
import os
import time

import celery
import redis.exceptions

from cubicweb.toolsutils import Command
from cubicweb.server.serverconfig import ServerConfiguration
from cubicweb.cwctl import CWCTL
from cubicweb_celerytask.entities import sync_task_state

from cw_celerytask_helpers.monitor import MONITOR_KEY
from cw_celerytask_helpers.utils import get_redis_client

logger = logging.getLogger(__name__)

REQUEUE_TIMEOUT = 600


class CeleryMonitorCommand(Command):
    """Synchronize celery task statuses"""

    name = 'celery-monitor'
    arguments = '<instance>'
    min_args = max_args = 1
    options = (
        ('loglevel',
         {'short': 'l', 'type': 'choice', 'metavar': '<log level>',
          'default': 'info', 'choices': ('debug', 'info', 'warning', 'error')},
         ),
    )

    def run(self, args):
        from cubicweb import repoapi
        from cubicweb.cwctl import init_cmdline_log_threshold
        config = ServerConfiguration.config_for(args[0])
        config.global_set_option('log-file', None)
        config.log_format = '%(levelname)s %(name)s: %(message)s'
        init_cmdline_log_threshold(config, self['loglevel'])
        # use only one database connection by default (require cubicweb>=3.24)
        os.environ.setdefault('CW_CONNECTIONS_POOL_SIZE', '1')
        repo = repoapi.get_repository(config=config)
        repo.hm.call_hooks('server_maintenance', repo=repo)
        source = repo.system_source
        while True:
            try:
                with repo.internal_cnx() as cnx:
                    while True:
                        try:
                            self.loop(cnx)
                        except redis.exceptions.ConnectionError:
                            logger.error('redis ConnectionError, retry in 5s')
                            time.sleep(5)
            except (source.OperationalError, source.InterfaceError):
                while True:
                    source.exception('connection lost, retry in 5s')
                    time.sleep(5)
                    try:
                        for cnxset in repo.cnxsets:
                            cnxset.reconnect()
                    except (source.OperationalError, source.InterfaceError):
                        continue
                    else:
                        break

    @staticmethod
    def requeue(cnx, client):
        """Requeue pending tasks in case we have missed some events.

        For instance, in case of "worker lost", we don't have task events.
        Only requeue tasks within the "visibility_timeout" (default 1h), since
        tasks status is lost after this period.
        """
        try:
            timeout = celery.current_app.conf[
                'BROKER_TRANSPORT_OPTIONS']['visibility_timeout']
        except KeyError:
            timeout = 3600
        date = datetime.datetime.utcnow() - datetime.timedelta(seconds=timeout)
        for task_id, in cnx.execute(
            'Any T ORDERBY C WHERE X is CeleryTask, X task_id T, '
            'X in_state ST, ST name IN ("waiting", "queued", "running"), '
            'X creation_date C, X creation_date > %(d)s', {'d': date},
            build_descr=False
        ):
            client.lpush(
                MONITOR_KEY,
                json.dumps({'task_id': task_id, 'task_name': None}))

    @staticmethod
    def loop(cnx, timeout=None):
        client = get_redis_client()
        client.ping()
        logger.info('Connected to redis')
        test = (cnx.repo.config.mode == "test")
        requeue_timer = timer = time.time()
        while True:
            # pop item from MONITOR_KEY
            data = client.brpop(MONITOR_KEY, timeout=1)
            if data is None:
                now = time.time()
                if timeout is not None and abs(timer - now) > timeout:
                    break
                if abs(requeue_timer - time.time()) > REQUEUE_TIMEOUT:
                    # no items left in MONITOR_KEY and we reached the
                    # REQUEUE_TIMEOUT, requeue tasks in MONITOR_KEY
                    CeleryMonitorCommand.requeue(cnx, client)
                    requeue_timer = time.time()
                continue
            payload = json.loads(data[1].decode())
            task_id, task_name = payload['task_id'], payload['task_name']
            try:
                sync_task_state(cnx, task_id, task_name)
            except Exception:
                logger.exception('Unhandled exception while syncing '
                                 'task <Task %s (%s)>', task_id,
                                 task_name)
                cnx.rollback()
                if test:
                    # we should not hide exceptions in tests
                    raise


CWCTL.register(CeleryMonitorCommand)



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-celerytask automatic tests base class"""

import os
import sys
import multiprocessing
import time

import redis
import celery
import celery.result
from celery.bin.worker import worker as celery_worker

from cubicweb.devtools import testlib
from cubicweb_celerytask.ccplugin import CeleryMonitorCommand


class BaseCeleryTaskTC(testlib.CubicWebTC):
    worker_args = ['-P', 'solo', '-c', '1', '-l', 'info']

    @classmethod
    def setUpClass(cls):
        super(BaseCeleryTaskTC, cls).setUpClass()
        REDIS_URL = os.environ.get('PIFPAF_REDIS_URL',
                                   'redis://localhost:6379/1')
        redis_client = redis.Redis.from_url(REDIS_URL)
        redis_client.flushall()
        task_module_path = cls.datapath('tasks')
        sys.path.insert(0, task_module_path)
        conf = celery.current_app.conf
        conf.broker_url = REDIS_URL
        conf.result_backend = REDIS_URL
        conf.CUBICWEB_CELERYTASK_REDIS_URL = REDIS_URL
        conf.task_always_eager = False
        conf.imports = ('cw_celerytask_helpers.helpers', 'tasks')
        # this is required since we use a non-cubicweb worker, so the startup
        # hook setting CUBICWEB_CELERYTASK_LOGDIR won't run.
        conf.CUBICWEB_CELERYTASK_LOGDIR = os.path.join(
            cls.config.appdatahome, 'logs')
        cls.worker = multiprocessing.Process(target=cls.start_worker)
        cls.worker.start()

    @classmethod
    def tearDownClass(cls):
        super(BaseCeleryTaskTC, cls).tearDownClass()
        cls.worker.terminate()
        cls.worker.join()

    @classmethod
    def start_worker(cls):
        app = celery.current_app
        worker = celery_worker(app)
        worker.run_from_argv("worker", cls.worker_args)

    def wait_async_task(self, cnx, task_id, timeout=5):
        result = celery.result.AsyncResult(task_id)
        start = time.time()
        while abs(time.time() - start) < timeout:
            if result.ready():
                CeleryMonitorCommand.loop(cnx, timeout=0)
                return result
            if not self.worker.is_alive():
                # will be joined in tearDown
                raise RuntimeError("Celery worker terminated")
            time.sleep(.1)
        raise RuntimeError("Timeout")



# -*- coding: utf-8 -*-
# copyright 2018 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os.path
import shutil

import celery
from cubicweb import ConfigurationError
from cubicweb.devtools import testlib
from cubicweb_celerytask.hooks import CeleryTaskStartupHook


class CeleryTaskStartupHookTC(testlib.CubicWebTC):

    def test_create_log_dir(self):
        """hook should set a celerytask-log-dir by default and create it"""
        logdir = self.config['celerytask-log-dir']
        self.assertEqual(
            self.config['celerytask-log-dir'],
            os.path.join(self.config.appdatahome, 'logs'))
        shutil.rmtree(logdir, ignore_errors=True)
        CeleryTaskStartupHook.setup_celerytask_logdir(self.config)
        self.assertTrue(os.path.isdir(logdir))

    def test_celeryconfig_log_dir(self):
        """when CUBICWEB_CELERYTASK_LOGDIR is set in celeryconfig and no
        celerytask-log-dir is set we must use value from
        CUBICWEB_CELERYTASK_LOGDIR"""
        conf = celery.current_app.conf
        logdir = os.path.join(self.config.appdatahome, 'testhook')
        old = (self.config['celerytask-log-dir'],
               conf['CUBICWEB_CELERYTASK_LOGDIR'])
        (self.config['celerytask-log-dir'],
         conf['CUBICWEB_CELERYTASK_LOGDIR']) = None, logdir
        try:
            CeleryTaskStartupHook.setup_celerytask_logdir(self.config)
            self.assertEqual(self.config['celerytask-log-dir'], logdir)
        finally:
            (self.config['celerytask-log-dir'],
             conf['CUBICWEB_CELERYTASK_LOGDIR']) = old
            os.rmdir(logdir)

    def test_mismatching_dirs(self):
        """hook should raise in case os mistmaching configuration"""
        conf = celery.current_app.conf
        old = conf['CUBICWEB_CELERYTASK_LOGDIR']
        conf['CUBICWEB_CELERYTASK_LOGDIR'] = '/some/dir'
        try:
            with self.assertRaises(ConfigurationError) as cm:
                CeleryTaskStartupHook.setup_celerytask_logdir(self.config)
            self.assertIn('You misconfigured your application',
                          str(cm.exception))
        finally:
            conf['CUBICWEB_CELERYTASK_LOGDIR'] = old



# pylint: disable=W0622
"""cubicweb-celerytask application packaging information"""


modname = 'cubicweb_celerytask'
distname = 'cubicweb-celerytask'

numversion = (0, 7, 1)
version = '.'.join(str(num) for num in numversion)

license = 'LGPL'
author = 'LOGILAB S.A. (Paris, FRANCE)'
author_email = 'contact@logilab.fr'
description = 'Run and monitor celery tasks'
web = 'http://www.cubicweb.org/project/%s' % distname

__depends__ = {
    'cubicweb': '>= 3.19',
    'six': '>= 1.4.0',
    'celery': '>=4,<5',
    'cw-celerytask-helpers': '>= 0.7.0, < 0.8.0',
}
__recommends__ = {}

classifiers = [
    'Environment :: Web Environment',
    'Framework :: CubicWeb',
    'Programming Language :: Python',
    'Programming Language :: JavaScript',
]



options = (
    ('celerytask-log-dir', {
        'type': 'string',
        'default': None,
        'help': ('celery task log directory; if unset, defaults to '
                 '<config.appdatahome>/logs'),
        'group': 'celerytask',
        'level': 2,
    }),
)
